/*************************************
  Little John GP32
  File : unes.c
  Authors : Yoyo.
  Version : 0.3
  Last update : 18th July 2003
**************************************/


/*#ifdef __testing__
#include "rom.h"
#endif*/

#define MAX_LOST_FRAME 5

#define __timerdiv__ 2

//#define __skipsynchro__

#include "gpdef.h"
#include "gpstdlib.h"
#include "gpgraphic.h"
#include "gpfont.h"
#include "gpmm.h"
#include "gpstdio.h"



#include "LJ_logo.h"
#include "gpfilereq.h"

#include "snss.h"
#include "font.h"
#include "unes.h"
#include "unes_ppu.h"
#include "unes_mapper.h"
#include "unes_io.h"
#include "nes_apu.h"
#include "unes_pal.h"
#include "ljohnGP.h"
#include "unes_crc32.h"

#include "mygraphic.h"

#include "unzip.h"
#include "unes_db.h"

#include "unes_tables.h"

#ifndef __asmcpu__
static nes6502_memread _NESread[] =
{
   /* $0 - $7FF is RAM */
   { 0x0800, 0xFFFF, Rd6502 },
   { -1,     -1,     NULL }
};
static nes6502_memwrite _NESwrite[] =
{
   /* $0 - $7FF is RAM */
   { 0x0800, 0xFFFF, Wr6502 },
   { -1,     -1,     NULL }
};
static nes6502_context _NEScontext;

#endif



// VNES
VirtualNES Vnes;

// Mapper
//extern Mapper_Info mapfunc;
mmc_t *mapper;
      

// GP32 stuff



int sz;
volatile int play_chan;
int *play_pos;
int gp32AudioPaused;

static uint16 key_pressed_delay;

GPDRAWSURFACE gpDraw[2];		// Priamry & Back SURFACE
GPDRAWTAG gpdrawtag;
int nflip;					// Flip Index
GP_HPALETTE h_nespal, h_menupal,h_orgpal;

// Emu stuff

byte vbl;
int romsize;
char gamename[256];
int fskiprate;
int fskiprate_adj,fskiprate_adj_ratio;
int frameskipped;
int fastforward;
long framecounter;
int frame,frameRendered;


bool framecompleted;
//bool vblank;
uint32  curscanline;
uint32  synccycles;
uint32  showFPS;
uint32 gamma;

uint32 stepbystep;
uint32 debugison;

volatile unsigned int framecount;

long beginTime;


char CurrentROMFile[256];
nestoy_info ninfo;
uint32 CYCLES_BEFORE_NMI;


pal normcolors[256],menucolors[256];
char chaine[1024];
char chaineDEB[1024];

uint32 scalePal[256];
uint32 scalePal2[256];


INLINE uint8 emulate_one_framePAL(void);
INLINE uint8 emulate_one_frameNTSC(void);

void nes_reset(void);
void CloseAll(void);
int SaveStateSnss(byte b);
int LoadStateSnss(byte b);

void Yo_Printc(uint32 x,uint32 y,char c,uint8 *scr,byte col);
void reset_asmcpu(void);

#include "unes_ROM_correct.c"

void timertps(void) {
	
	framecount++;
}

long gp32_GpTickCountGet(void)
{
	return GpTickCountGet()/__timerdiv__;
}

void gpdelay(long l)
{
   long tt;
	
  tt=gp32_GpTickCountGet();
  while (gp32_GpTickCountGet()-tt<l)
  {
    __asm {nop;nop;nop;nop;}
  }

}


void gpwait(int l)
{
   int i;
   long tt;
	
   if (l<=0) 
   {
   	#ifndef __fastkey__
   	while (1)
  	{
  	  GpKeyGetEx(&i);
  	  //if (i & GPC_VK_START) return;
  	  if (!i) break;  	  
	}		
	#endif
	#ifndef __fastkey2__
	while (1)
  	{
  	  GpKeyGetEx(&i);
  	  //if (i & GPC_VK_START) return;
  	  if (i) break;  	  
	} 	
	if ((i&GPC_VK_RIGHT)||(i&GPC_VK_LEFT)) 
	{
		stepbystep=1;	
		debugison=1;
	}
	else 
	{
		stepbystep=0;	
		debugison=0;
}		
	if (i&GPC_VK_LEFT) return;
	#endif
	#ifndef __fastkey__
	while (1)
  	{
  	  GpKeyGetEx(&i);
  	  //if (i & GPC_VK_START) return;
  	  if (!i) break;  	  
	} 
	#endif
	
	#ifdef __fastkey3__	
	l=1;
	#else
	return;
	#endif
	
   }
	
  tt=gp32_GpTickCountGet();
  while (gp32_GpTickCountGet()-tt<l*100)
  {
  	i=1;
    __asm {nop;nop;nop;nop;}
  }

}


void gpprintf(int x,int y,char *chaine,int pause,int bg,int col)
{  
  int i=0;
  int j=1;
  int maxl=0;
  int debl=0;
  int endl=0;
  if (!chaine) return;
  
  while (chaine[i])
  {
  	if (chaine[i]=='\n') 
  	{
  		j++;
  		if (endl-debl>maxl) maxl=endl-debl;
  		endl=debl=i;
  	}
  	endl++;
  	i++;
  }
  if (endl-debl>maxl) maxl=endl-debl;  
  if (!maxl) maxl=i;
  if (bg!=-1) GpRectFill(NULL,&gpDraw[nflip],x,y,maxl*8,j*16,bg);  
  GpTextOut(NULL,&gpDraw[nflip],x,y,(char*)chaine,col);
  if (pause!=-1) gpwait(pause);
}

void GP_PauseAudio(Uint8 pause)
{    		       
  if (pause==1)
    {
      if (Vnes.var.enablesound==2) GpPcmStop();
      gp32AudioPaused=1;
    }
  else
    {
      gp32AudioPaused=0;
      if (Vnes.var.enablesound==2)
      {
      	GpPcmPlay((unsigned short*)Vnes.var.sndbuffer, Vnes.var.buffsize*NBBUFFERS<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo),1);
      	GpPcmLock((unsigned short*)Vnes.var.sndbuffer,(int *)&play_chan,(unsigned int *)&play_pos);
      }
      //put it to write in the middle...
      Vnes.var.sndbuffer_pos=Vnes.var.sndbuffer+ (Vnes.var.buffsize*NBBUFFERS<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo))/2;       
      Vnes.var.sndbuffer_end=Vnes.var.sndbuffer+ (Vnes.var.buffsize*NBBUFFERS<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo));      
      //size of sound buffer in bytes
      sz=Vnes.var.buffsize*NBBUFFERS<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo); 
    }
}

void GP_SetPal(int a)
{	
  switch (a)
    {
    case 0:
      //palette change
      GpPaletteDelete(GpPaletteSelect(h_nespal));				//previous palette pointer (h_pal) will be lost
      GpPaletteRealize();
      break;
    case 1:
      GpPaletteDelete(GpPaletteSelect(h_menupal));				//previous palette pointer (h_pal) will be lost
      GpPaletteRealize();
      break;	
    case 2:
      GpPaletteDelete(GpPaletteSelect(h_orgpal));				//previous palette pointer (h_pal) will be lost
      GpPaletteRealize();
      break;	
      
    }
}


void GP_ReSetPalette()
{  
  int i;
  GP_PALETTEENTRY*	tmp_entry;
  uint32 rgamma[256],ggamma[256],bgamma[256];
      
  for (i=0;i<256;i++)
  {  	
  	rgamma[i]=gammatab[gamma][i];
  	ggamma[i]=gammatab[gamma][i];
  	bgamma[i]=gammatab[gamma][i];  	
  }

  tmp_entry = (GP_PALETTEENTRY*)gm_malloc((sizeof(GP_PALETTEENTRY) * 256));
  for(i = 0 ; i < 255 ; i++)
    {
      tmp_entry[i] = ((rgamma[normcolors[i].r]/8)<<11) | ((ggamma[normcolors[i].g]/8)<<6) | ((bgamma[normcolors[i].b]/8)<<1) | 0;	//if LSB is cleared, the color is more clear
    }
  GpPaletteDelete(h_nespal);
  h_nespal = GpPaletteCreate(256, tmp_entry);
  if ( !h_nespal )
    {
      //error
      
    }
  
  for(i = 0 ; i < 255 ; i++) 
    {
      tmp_entry[i] = ((rgamma[menucolors[i].r]/8)<<11) | ((ggamma[menucolors[i].g]/8)<<6) | ((bgamma[menucolors[i].b]/8)<<1) | 0;	//if LSB is cleared, the color is more clear
    }
  GpPaletteDelete(h_menupal);
  h_menupal = GpPaletteCreate(256, tmp_entry);
  
  gm_free(tmp_entry);  
}


void GP_SetPalette()
{
  
  int i;
  GP_PALETTEENTRY*	tmp_entry;
  uint32 rgamma[256],ggamma[256],bgamma[256];
  for (i=0;i<256;i++)
  {
  	rgamma[i]=gammatab[gamma][i];
  	ggamma[i]=gammatab[gamma][i];
  	bgamma[i]=gammatab[gamma][i];  	
  }

  tmp_entry = (GP_PALETTEENTRY*)gm_malloc((sizeof(GP_PALETTEENTRY) * 256));
  for(i = 0 ; i < 255 ; i++) 
    {
      tmp_entry[i] = ((rgamma[normcolors[i].r]/8)<<11) | ((ggamma[normcolors[i].g]/8)<<6) | ((bgamma[normcolors[i].b]/8)<<1) | 0;	//if LSB is cleared, the color is more clear
      scalePal[i] = ((rgamma[normcolors[i].r]/8)<<11) | ((ggamma[normcolors[i].g]/8)<<6) | ((bgamma[normcolors[i].b]/8)<<1) | 0;	//if LSB is cleared, the color is more clear
      scalePal2[i] = ((rgamma[normcolors[i].r]/8)<<16) | ((ggamma[normcolors[i].g]/8)<<8) | ((bgamma[normcolors[i].b]/8)<<0) | 0;
    }
  h_nespal = GpPaletteCreate(256, tmp_entry);
  if ( !h_nespal )
    {
      //error
      
    }
  
  for(i = 0 ; i < 255 ; i++) 
    {
      tmp_entry[i] = ((rgamma[menucolors[i].r]/8)<<11) | ((ggamma[menucolors[i].g]/8)<<6) | ((bgamma[menucolors[i].b]/8)<<1) | 0;	//if LSB is cleared, the color is more clear
    }
  h_menupal = GpPaletteCreate(256, tmp_entry);
  if ( !h_menupal )
    {
      //error
    }
  gm_free(tmp_entry);
  //Get original system pal & set menupalette
  h_orgpal = GpPaletteSelect(h_menupal);
}

void Message (int org_x,int org_y,const char *message,byte *buff,byte col)
{
  int l;
  int k;
  int msg_x,msg_y;

  
  if (message)
  {
      l=gm_lstrlen(message);
      msg_x=org_x;
      msg_y=org_y;
      k=0;   
      while (k<l)
	{
	  if (message[k]!='\n') 
	    {
	      Yo_Printc(msg_x,msg_y,message[k],(uint8*)buff,col);
	      msg_y+=8;
	    }
	  else 
	    {
	      msg_x+=6;
	      msg_y=org_y;
	    }
	  if (msg_y>=(256-8))
	    {
	      msg_x+=6;
	      msg_y=0;
	    }
	  if (msg_x>=(320-6))
	    {
	      msg_x=0;
	      msg_y=0;	      
	    }
	  k++;
	}
    }
  
}


#include "unes_gui.c"
char Init_NES(char *RomName)
{
  int i,j,k;
  int err_code;
  long l;
  byte b;
  char sramfile[256];
  char strresult[256];
  ulong bread;
  //  FILE *f;
  F_HANDLE f;
  
  gm_memset(Vnes.PPU_patterntables,0,0x8000);
  gm_memset(Vnes.PPU_nametables,0,0x1000);
  gm_memset(Vnes.PPU_palette,0,0x100);
  
  gm_memset(Vnes.CPUMemory,0,65536);  
  
  gm_memset(Vnes.NESSRAM,0,0x10000);
  gm_memset(Vnes.mapper_extram,0,0x10000);
  
  Vnes.sramsize=0x2000;
  Vnes.vramsize=0x2000;
  
    
  //and the mirror, according to the method choosen, 4screen not implemented yet!!!!
  if (Vnes.ROM_Header.rom_ctrl1&1)
    {
      //Vertical Vnes.var.mirroring : name table 0,1 are orginals, 2,3 are mirrors
      Vnes.var.mirroring=1;      
    }
  else
    {
      //Horizontal Vnes.var.mirroring : name table 0,2 are orginals, 1,3 are mirrors
      Vnes.var.mirroring=2;
    }
  if (Vnes.ROM_Header.rom_ctrl1&8) Vnes.var.mirroring=5;    
   
  
  
  
  //set vram 
  Vnes.var.vrom=0;
  
  //set low cpu page
  Vnes.CPUPageIndex[0]=Vnes.CPUMemory;
  Vnes.CPUPageIndex[1]=Vnes.CPUMemory+8192;
  Vnes.CPUPageIndex[2]=Vnes.CPUMemory+16384;  
  Vnes.CPUPageIndex[3]=Vnes.NESSRAM; //Vnes.CPUMemory+16384+8192;
  
  if (Vnes.trainer) gm_memcpy((uint8*)(&Vnes.CPUPageIndex[3][0x1000]),(uint8*)(Vnes.rom+16),512);

    
  //  printf("  * MMC Init...");

    
  //if (InitMapper()) return 1;  
  mapper=(mmc_t *)mmc_create();
  if (!mapper) 
  {
  	gm_sprintf(chaine,"Mapper %d not supported\nPress a button",Vnes.Mapper_used);
  	DrawMessage(chaine,nflip);
  	gpwait(0);
  	return 1;  
  }
      
  ppu_init();      
    
  Init_Sound();
  //  printf("ok\n");
  // Init the CPU
  
  
    
  #ifndef __asmcpu__        
  nes6502_getcontext(&_NEScontext);
  _NEScontext.mem_page[0]=(uint8 *)Vnes.CPUPageIndex[0];
  /*_NEScontext.mem_page[1]=(uint8 *)Vnes.CPUPageIndex[1];
  _NEScontext.mem_page[2]=(uint8 *)Vnes.CPUPageIndex[2];
  _NEScontext.mem_page[3]=(uint8 *)Vnes.CPUPageIndex[3];
  _NEScontext.mem_page[4]=(uint8 *)Vnes.CPUPageIndex[4];
  _NEScontext.mem_page[5]=(uint8 *)Vnes.CPUPageIndex[5];
  _NEScontext.mem_page[6]=(uint8 *)Vnes.CPUPageIndex[6];
  _NEScontext.mem_page[7]=(uint8 *)Vnes.CPUPageIndex[7];*/
  _NEScontext.read_handler=_NESread;
  _NEScontext.write_handler=_NESwrite;        
  nes6502_setcontext(&_NEScontext);    
  #endif   
    
  Vnes.var.DrawCframe=1;
  frameskipped=0;
	  	  
  Vnes.var.JoyPad1_BitIndex=0;
  Vnes.var.JoyPad2_BitIndex=0;	  
	
  Vnes.var.SaveRAM=Vnes.ROM_Header.rom_ctrl1&2;  
  fastforward=0;        
  //  printf("ok\n");
	
  //	if (Vnes.ROM_Header.rom_ctrl1&2)
  {
    //load the sram
    i=0;
    j=255;
    k=-1;
    while (RomName[i]!=0) 
      {
	if (RomName[i]=='.') j=i;
	if (RomName[i]=='/') k=i;
	if (RomName[i]=='\\') k=i;
	i++;
      }
    if (j==255) j=i;  //pas d'extension	  
    gm_memcpy(sramfile,RomName+k+1,j+1-k-1);    
    sramfile[j+1-k-1]=0;
    gm_strcat(sramfile,"srm");        
//    gm_memcpy(result,path_srm,gm_lstrlen(path_srm)+1);
    //gm_sprintf(strresult,"%s%s",path_srm,sramfile);
//    gm_strcpy(result,path_srm);    
//    gm_strcat(result,sramfile);	  
    err_code = GpFileOpen(sramfile, OPEN_R, &f);
    if (err_code != SM_OK)
    {
	//gpprintf(0,48,"No sram data",5);
    }
    else
    {
	err_code = GpFileRead(f, Vnes.NESSRAM,0x2000,&bread);
	if (err_code != SM_OK)
	{
	    //gpprintf(0,48,"SRAM loading error",5);
	    //DrawMessage("SRAM loading error",nflip);
	    //gpwait(5);
	}
	else
	{
	    //gpprintf(0,48,"SRAM loaded",5);
	    //DrawMessage("SRAM loaded",nflip);
	    //gpwait(5);
	    Vnes.var.SaveRAM=1;
	}
	GpFileClose(f);
    }
  }  
  return 0;
}

char Load_ROM(char *RomName)
{
  //FILE *f;
  F_HANDLE f;
  unsigned long filesize;
  int i,j,k;
  ulong readsize;
  char ext[5];
  char strfsize[256],strfname[256];
  char c[256];
  int	err_code;
  unzFile zip_file = 0;    
  char nes_file[256]; //yoyo
  char *p;
  unz_file_info unzinfo;
  char *pext;
  
  gm_sprintf(strfsize,"Loading & analyzing...");//,romsize);
  DrawMessage(strfsize,nflip);
	

  pext = (char*)yo_strrchr(RomName, '.');
  if (pext) gm_strncpy(ext, pext, sizeof(ext));
  yo_strlwr(ext);
  if (gm_compare(ext, ".zip") == 0)  
  {
  	zip_file = unzOpen(RomName);
  	
			  

                            
        if (!zip_file) return 1;
        
                                
        unzGoToFirstFile (zip_file);
        
            
        for (;;) 
        {
        	if (unzGetCurrentFileInfo(zip_file, &unzinfo, nes_file, sizeof(nes_file), NULL, NULL, NULL, NULL) != UNZ_OK) return 1;

         	    yo_strlwr(nes_file); 

                p = (char *)yo_strrchr(nes_file, '.');
                                
                if (gm_compare(p, ".nes") == 0) break;
                if (gm_compare(p, ".nsf") == 0) break;
                if (unzGoToNextFile(zip_file) != UNZ_OK) return 1;
         }         
         unzOpenCurrentFile (zip_file);

	 romsize = unzinfo.uncompressed_size;	 
	 
	 
	 Vnes.rom=(uint8 *)gm_malloc(romsize);
	 unzReadCurrentFile(zip_file, Vnes.rom, romsize);
         //if(unzReadCurrentFile(zip_file, Vnes.rom, romsize) != romsize) return 1;
         unzCloseCurrentFile (zip_file);
         unzClose (zip_file);
                
  }
  else
  {
  
  	err_code = GpFileGetSize(RomName, &filesize);
  	if (err_code != SM_OK)	
  	{  
  		GpTextOut(NULL,&gpDraw[nflip],160,10,(char*)"File access error",0xe0);
  		return;
  	}
  
  	romsize=filesize;          
  
  	err_code = GpFileOpen(RomName, OPEN_R, &f);
  	if ((err_code != SM_OK)||(romsize==0))
  	{
  		GpTextOut(NULL,&gpDraw[nflip],10,30,(char*)"Could not find romfile",0xe0);  		
  		Vnes.rom=NULL;
  		GpFileClose(f);
  		return(1);
  	}
	
  	Vnes.rom=(uint8 *)gm_malloc(romsize);
  	if (!Vnes.rom)
  	{  		
  		GpFileClose(f);
  		return(2);
  	}

  	err_code = GpFileRead(f, Vnes.rom, romsize, &readsize);

	if (err_code != SM_OK)
  	{
  		GpFileClose(f);
  		return(3);
  	}
  	GpFileClose(f);  	
  }

  
  return 0;
}

char Open_ROM()
{
  int i,j,x,y;
  word addr;
  
  Vnes.romsize=romsize;      
  
  if ((Vnes.rom[0]!='N')||(Vnes.rom[1]!='E')||(Vnes.rom[2]!='S'))
    {      
      return(4);
    }    

  if (Vnes.rom[3]=='M')
    {//test for a .nsf file    	      
      Vnes.var.nsfmode=1;
      return 1;
    }     
  else Vnes.var.nsfmode=0;
  
  gm_strcpy(gamename,CurrentROMFile);  
  if (Vnes.var.nsfmode)
  {  	          	  
  	
  } 

  
  gm_memcpy(&Vnes.ROM_Header,Vnes.rom,16);  	
  Vnes.trainer = (Vnes.ROM_Header.rom_ctrl1&0x4?1:0);      
  

  /*compute crc*/
  Vnes.var.crc = 0;
  Vnes.var.crc_all = 0;  
  Vnes.var.country[0]='?';
  Vnes.var.country[1]=0;
  Vnes.var.publisher[0]='?';
  Vnes.var.publisher[1]=0;
  Vnes.var.date[0]='?';
  Vnes.var.date[1]=0;  
  if (!Vnes.var.nsfmode)
  {

  	Vnes.var.crc = CrcCalc(Vnes.rom+16, Vnes.trainer*512+Vnes.ROM_Header.prg_rom_pages_nb * 0x4000);//, Vnes.var.crc);
  	Vnes.var.crc_all = CrcCalca(Vnes.var.chr_beg, Vnes.ROM_Header.chr_rom_pages_nb * 0x2000, Vnes.var.crc);     

  	
  	if (!db_getinfo(Vnes.var.crc,Vnes.var.crc_all,&ninfo))
  	{
	  	gm_strcpy(gamename,ninfo.title);
	  	gm_strcpy(Vnes.var.country,ninfo.country);
	  	gm_strcpy(Vnes.var.publisher,ninfo.publisher);
	  	gm_strcpy(Vnes.var.date,ninfo.date);  
  		if ((Vnes.var.country[0]=='E')||(Vnes.var.country[0]=='S'))
 		{
	  		Vnes.var.fps=50;
	  	}
	  	Vnes.ROM_Header.rom_ctrl1=ninfo.header1;
	  	Vnes.ROM_Header.rom_ctrl2=ninfo.header2;
  	}  	

  	Crc_CorrectRomSettings();
  	if (Vnes.rom[12]==0)
    	  Vnes.Mapper_used=(Vnes.ROM_Header.rom_ctrl1>>4)+(Vnes.ROM_Header.rom_ctrl2&0xf0);
  	else
    	  Vnes.Mapper_used=(Vnes.ROM_Header.rom_ctrl1>>4);    
 	Vnes.var.chr_beg=Vnes.rom+16+Vnes.trainer*512+Vnes.ROM_Header.prg_rom_pages_nb*16384;
  	Vnes.var.prg_beg=Vnes.rom+16+Vnes.trainer*512;
  	Vnes.var.cart_size=(romsize-16-Vnes.trainer*512)/1024;
  }
  

  
  Vnes.CPUMemory=(uint8 *)gm_malloc(65536);
  Vnes.PPU_patterntables=(uint8 *)gm_malloc(0x8000);
  Vnes.PPU_nametables=(uint8 *)gm_malloc(0x1000);
  Vnes.PPU_palette=(uint8*)gm_malloc(0x100);
  
  Vnes.mapper_extram=(uint8 *)gm_malloc(0x10000);  
  Vnes.mapper_extramsize=0;
  Vnes.NESSRAM=(uint8*)gm_malloc(0x10000);
  
  if ( (!Vnes.CPUMemory)||(!Vnes.PPU_patterntables)||(!Vnes.PPU_nametables)||(!Vnes.PPU_palette)||
  	(!Vnes.mapper_extram)||(!Vnes.NESSRAM))
    {      
      return(5);
    }

  // PPU init
  
  return(0);
}

char Close_ROM(int b)
{
  //  FILE *f;
  F_HANDLE f;  
  int i,j,k;
  int err_code;
  char sramfile[256],result[256];
  if (b)
    {
      if (Vnes.var.SaveRAM)
	{
	  //save the sram
	  i=0;
	  j=255;
	  k=-1;
	  while (CurrentROMFile[i]!=0) 
	    {
	      if (CurrentROMFile[i]=='.') j=i;
	      if (CurrentROMFile[i]=='/') k=i;
	      if (CurrentROMFile[i]=='\\') k=i;
	      i++;
	    }
	  if (j==255) j=i;  //pas d'extension
	  gm_memcpy(sramfile,CurrentROMFile+k+1,j+1-k-1);
	  sramfile[j+1-k-1]=0;
	  gm_strcat(sramfile,"srm");
	  	  
	  //gm_memcpy(result,path_srm,gm_lstrlen(path_srm)+1);
	  //gm_strcat(result,sramfile);
	  
	  //err_code = GpFileRemove(sramfile);
	  //gpwait(10);
	  err_code = GpFileCreate(sramfile, ALWAYS_CREATE, &f);
	  if (err_code != SM_OK)
	  {
	  	//gpprintf(0,48,"Error during creation",5);
	  	DrawMessage("Error during creation",nflip);
	  	gpwait(5);
	  }
	  else
	  {
	  	err_code = GpFileWrite(f, Vnes.NESSRAM ,0x2000);
	   	if (err_code != SM_OK)
	  	{
	  	    //gpprintf(0,48,"Error during saving",5);
	  	    DrawMessage("Error while writing file",nflip);
	  	    gpwait(5);
	    	}
	    	GpFileClose(f);
	    	//gpprintf(0,48,"File successfully saved",5);
	    	//Message("File successfully saved");
	  	//gpwait(5);
	  }
	}
    }
    
  mmc_destroy(&mapper);
  
  GpPaletteDelete(h_menupal);
  GpPaletteDelete(h_nespal);
  GpPaletteDelete(h_orgpal);
  
  if (Vnes.rom) gm_free(Vnes.rom);
  if (Vnes.CPUMemory) gm_free(Vnes.CPUMemory);
  if (Vnes.PPU_patterntables) gm_free(Vnes.PPU_patterntables);
  if (Vnes.PPU_nametables) gm_free(Vnes.PPU_nametables);
  if (Vnes.PPU_palette) gm_free(Vnes.PPU_palette);
  if (Vnes.mapper_extram) gm_free(Vnes.mapper_extram);
  if (Vnes.NESSRAM) gm_free(Vnes.NESSRAM);    
  if (Vnes.var.Vbuffer) gm_free(Vnes.var.Vbuffer);
  

  return(0);
}

void nes_reset()
{  
  Vnes.var.sramEnabled=Vnes.var.SaveRAM;	
	
  ppu_reset(true);    
  apu_reset();  
  mmc_reset();
  
  if (ppu_noLatchFunc()) Vnes.var.lazy_mode_avail=1;
  else Vnes.var.lazy_mode_avail=0;
  
  if (Vnes.var.lazy_mode_avail) Vnes.var.lazy_mode=1;
  else Vnes.var.lazy_mode=0;
  
  Vnes.var.frame_irq_enabled = 0xFF;
  Vnes.var.frame_irq_disenabled = 0;
    
  #ifdef __asmcpu__
  reset_asmcpu();
  #else
  nes6502_reset();  
  #endif  
          
  beginTime=gp32_GpTickCountGet();  
  framecount=0;
  framecounter=0;  
  key_pressed_delay=0;
  
  Vnes.var.curdeccycle=0;
  Vnes.var.currentcpucycle=0;
}

void LaunchEmu()
{
  long l;
  int i,j,fx;
  char chaine[512];      
  
  GP_SetPal(2);
  Cls(0,nflip);
  /*gm_strcpy(Vnes.var.publisher,"Unknown");
  gm_strcpy(Vnes.var.country,"xxx");*/
  
  
  
       
  gm_sprintf(chaine,"[LJGP32 0.3]\nBrought to you by Yoyo\n\n%s\nSize : %dKo P-%dKo V-%dKo\nMapper : %d\n\n[ENJOY! ^=^]\nPublisher = %s Country = %s\nSpeed 133Mhz    Mem : %d\n\n Press a button to start!",
    gamename,Vnes.var.cart_size,Vnes.ROM_Header.prg_rom_pages_nb*16,Vnes.ROM_Header.chr_rom_pages_nb*8,Vnes.Mapper_used,
    Vnes.var.publisher,Vnes.var.country,gm_availablesize());
  DrawMessage(chaine, nflip);  
  gpwait(0);
  framecount=0;
  framecounter=0;
  frame=0;
  l=16;
  srand(gp32_GpTickCountGet());
  fx=rand()%3;
  do
  {
  	framecounter++;  	
  	l--;
  	while (framecount<framecounter)
  	{
  	    l=l+1;
  	    l=l-1;
  		__asm {nop;nop;nop;nop;}
  	}
  	switch (fx)
  	{
  		case 0:
  		for (i=0;i<320;i++)
    		for (j=0;j<15;j++) 
    		{
		    	(&gpDraw[nflip])->ptbuffer[(i*240+(239-(j<<4|l)))]=0;
	    	}  
	    	break;
	    	case 1:
	    	for (i=0;i<20;i++)
    		for (j=0;j<240;j++) 
    		{
		    	(&gpDraw[nflip])->ptbuffer[((i<<4|l)*240+(239-j))]=0;
	    	}  
	    	break;
	    	case 2:
	    	for (i=0;i<320;i++)
    		for (j=0;j<240;j++) 
    		{
		    	if (((i+j)&0xf)==l) (&gpDraw[nflip])->ptbuffer[(i*240+(239-j))]=0;
	    	}  
	    	break;
	    }
  	
  }
  while (l);
  
  
  //GpTextOut(NULL,&gpDraw[nflip],0,0,(char*)chaine,1);
  //gpwait(2);  
  //gpprintf(0,0,chaine,10);
    
  GP_PauseAudio(1);
  nes_reset();          
  GP_SetPal(0);
  Cls(13,0);
  Cls(13,1);
  GP_PauseAudio(0);

	
  if (Vnes.var.fps==60)
  {
  	while (1)
  	{     	
      		if (emulate_one_frameNTSC()) break;      
    	}
  }
  else
  {
  	while (1)
  	{     	
      		if (emulate_one_framePAL()) break;      
    	}
  }
  
  ppu_deinit();
}

long JoyPad1_State()
{
  long state; 
  int ExKey;  
  state=0;
        
  if ((Vnes.var.padmode==0)||(Vnes.var.padmode==2)) 
  {
  	GpKeyGetEx(&ExKey);    
  	if (Vnes.var.padmode==0)
  	{
  		if(ExKey & GPC_VK_LEFT) state|=0x40;
  		if(ExKey & GPC_VK_UP) state|=0x10;
  		if(ExKey & GPC_VK_RIGHT) state|=0x80;
  		if(ExKey & GPC_VK_DOWN) state|=0x20;
  		if(ExKey & GPC_VK_FB)  
  		{  	  		
  			switch (Vnes.var.autofireB)
  			{
  				case 0:state|=2; break;
  				case 1:
  		   		if ((framecount&31)>15) state|=2;
  		   		break;
  				case 2:
  		   		if ((framecount&15)>7) state|=2;
  		   		break;
  				case 3:
  		   		if ((framecount&7)>3) state|=2;
  		   		break;  		   
			}  		
  		}
  		if(ExKey & GPC_VK_FA)  
  		{  	  		
  			switch (Vnes.var.autofireA)
  			{
  				case 0:state|=1; break;
  				case 1:
  		   		if ((framecount&31)>15) state|=1;
  		   		break;
  				case 2:	
  		   		if ((framecount&15)>7) state|=1;
  		   		break;
  				case 3:
  		   		if ((framecount&7)>3) state|=1;
  		   		break;  		   
			}  		
  		}
  	}
  	if(ExKey & GPC_VK_START) state|=8;
  	if(ExKey & GPC_VK_SELECT) state|=4;
  }  	
  state|=0x10000;			//1player signature  	  
  return(state);
}

long JoyPad2_State()
{
  long state; 
  int ExKey;  
      
  if (Vnes.var.padmode!=1) return 0x20000;
  GpKeyGetEx(&ExKey);
  state=0;
      
  if(ExKey & GPC_VK_LEFT) state|=0x40;
  if(ExKey & GPC_VK_UP) state|=0x10;
  if(ExKey & GPC_VK_RIGHT) state|=0x80;
  if(ExKey & GPC_VK_DOWN) state|=0x20;
  if(ExKey & GPC_VK_FB)  
  {  	  		
  	switch (Vnes.var.autofireB)
  	{
  		case 0:state|=2; break;
  		case 1:
  		   if ((framecount&31)>15) state|=2;
  		   break;
  		case 2:
  		   if ((framecount&15)>7) state|=2;
  		   break;
  		case 3:
  		   if ((framecount&7)>3) state|=2;
  		   break;  		   
	}  		
  }
  
  if(ExKey & GPC_VK_FA)  
  {  	  		
  	switch (Vnes.var.autofireA)
  	{
  		case 0:state|=1; break;
  		case 1:
  		   if ((framecount&31)>15) state|=1;
  		   break;
  		case 2:
  		   if ((framecount&15)>7) state|=1;
  		   break;
  		case 3:
  		   if ((framecount&7)>3) state|=1;
  		   break;  		   
	}  		
  }
  if(ExKey & GPC_VK_START) state|=8;
  if(ExKey & GPC_VK_SELECT) state|=4;
  state|=0x20000;			//1player signature  
  return(state);
}

int SaveStateSnss(byte b)
{	
  SNSS_FILE *f;
  int err_code;
  char savename[255],result[256];
  byte i=0,j=255,k=255,l=255;
  char chaine[256];
#ifdef __asmcpu__
  uint32 nesregs[6];
#endif

  SNSS_RETURN_CODE code;
       
  while (CurrentROMFile[i]) 
    {
      if (CurrentROMFile[i]==92) j=i;  //92="\" 
      if (CurrentROMFile[i]==46) k=i;  //46="."
      if (CurrentROMFile[i]=='/') l=i; 
      i++;
    }	
  CurrentROMFile[k]=0;
  if (l==255) gm_strcpy(savename,&CurrentROMFile[(byte)(j+1)]);  
  else gm_strcpy(savename,&CurrentROMFile[(byte)(l+1)]);  
  CurrentROMFile[k]=46;
  i=gm_lstrlen(savename);
  savename[i]=46;
  savename[i+1]='s';
  savename[i+2]='s';
  savename[i+3]=b%10+48;
  savename[i+4]=0;
  gm_sprintf(chaine,"Saving state to file : %s\n",savename);
  gpprintf(0,16,chaine,2,-1,13);
  /*  f=fopen(result,"wb");
      if (f==NULL)
      {
      DrawMessage("Cannot save state\n",nflip);
      return 1;
      }*/
  //f = openSnssFile (savename,SNSS_OPEN_WRITE);  
  code = SNSS_OpenFile (&f,savename,SNSS_OPEN_WRITE);
  /*err_code = GpFileCreate(savename, ALWAYS_CREATE, &f);
  if (err_code != SM_OK)*/
  if (code != SNSS_OK)
  {
	gpprintf(0,48,"Error during creation",5,-1,13);
	return 1;
  }  
    
#ifndef __asmcpu__  
  nes6502_getcontext(&_NEScontext);  
  f->baseBlock.regA=_NEScontext.a_reg;
  f->baseBlock.regX=_NEScontext.x_reg;
  f->baseBlock.regY=_NEScontext.y_reg;
  f->baseBlock.regFlags=_NEScontext.p_reg;
  f->baseBlock.regStack=_NEScontext.s_reg;
  f->baseBlock.regPc=_NEScontext.pc_reg;  
#else
  get_context(nesregs);
  f->baseBlock.regA=nesregs[0];
  f->baseBlock.regX=nesregs[1];
  f->baseBlock.regY=nesregs[2];
  f->baseBlock.regFlags=nesregs[3];
  f->baseBlock.regStack=nesregs[5];
  f->baseBlock.regPc=nesregs[4]; 
#endif
  gm_memcpy(f->baseBlock.cpuRam,Vnes.CPUPageIndex[0],0x800);  
  ppu_savestate(&(f->baseBlock));
    
  gm_memcpy(f->baseBlock.ppuRam,Vnes.PPU_nametables,0x1000);
  gm_memcpy(f->baseBlock.palette,Vnes.PPU_palette,0x20);

  for (i=0;i<0x20;i++)
    {
      f->baseBlock.palette[i]&=0x3f;
    }
	
  f->baseBlock.mirrorState[0]=Vnes.var.ppu_mirror0;
  f->baseBlock.mirrorState[1]=Vnes.var.ppu_mirror1;
  f->baseBlock.mirrorState[2]=Vnes.var.ppu_mirror2;
  f->baseBlock.mirrorState[3]=Vnes.var.ppu_mirror3;  
  
  //  printf ("writing Base block\n");	  	  
  if ((code = SNSS_WriteBlock(f,SNSS_BASR)) != SNSS_OK)
    {
      //      printf ("Error: %s\n", getSnssErrorString (code));
      return 1;
    }
    gm_sprintf(chaine,"Wrote baseblock");
  gpprintf (0,16*3,chaine,-1,-1,13);

  f->vramBlock.vramSize=Vnes.vramsize;
  gm_memcpy(f->vramBlock.vram,Vnes.PPU_patterntables,f->vramBlock.vramSize);
  //  printf ("writing VRAM block\n");
  if ((code = SNSS_WriteBlock(f,SNSS_VRAM)) != SNSS_OK)
    {
      //      printf ("Error: %s\n", getSnssErrorString (code));
      return 1;
    }
  gpprintf (0,16*4,"Wrote VRAM block",-1,-1,13);
  f->sramBlock.sramSize=Vnes.sramsize;
  f->sramBlock.sramEnabled=Vnes.var.sramEnabled;
  gm_memcpy(f->sramBlock.sram,Vnes.NESSRAM ,f->sramBlock.sramSize);
  //  printf ("writing SRAM block\n");
  
  if ((code = SNSS_WriteBlock(f,SNSS_SRAM)) != SNSS_OK)
    {
      //      printf ("Error: %s\n", getSnssErrorString (code));
      return 1;
    }
  gpprintf (0,16*5,"Wrote SRAM block",-1,-1,13);

  for (i = 0; i < 4; i ++)
    {
      f->mapperBlock.prgPages[i]=(Vnes.CPUPageIndex[i+4]-Vnes.var.prg_beg) >> 13;
    }

  for (i = 0; i < 8; i ++)
    {
      if (Vnes.PPUPageIndexProtect[i])
        f->mapperBlock.chrPages[i]=(Vnes.PPUPageIndex[i]-Vnes.var.chr_beg) >> 10;
      else
        f->mapperBlock.chrPages[i]=((Vnes.PPUPageIndex[i]-Vnes.PPU_patterntables) >> 10) | 0x8000;
    }
  
  if (mapper->intf->get_state) mapper->intf->get_state(&(f->mapperBlock));


  //   printf ("writing mapper data block\n");
  if ((code = SNSS_WriteBlock(f,SNSS_MPRD)) != SNSS_OK)
    {
      //      printf ("Error: %s\n", getSnssErrorString (code));
      return 1;
    }
  gpprintf (0,16*6,"Wrote mapper block",-1,-1,13);
    
  
  gm_memcpy(f->soundBlock.soundRegisters,&Vnes.CPUMemory[0x4000],0x16);
  SNSSSaveStateSound(&(f->soundBlock));  
  if ((code = SNSS_WriteBlock(f,SNSS_SOUN)) != SNSS_OK)
	{
	  //	  printf ("Error: %s\n", getSnssErrorString (code));
	  return 1;
	}
  gpprintf (0,16*8,"Wrote sound block",-1,-1,13);

  /* close the files */
  if ((code=SNSS_CloseFile(&f)) != SNSS_OK)
    {
      //      printf ("Error: %s\n", getSnssErrorString (code));
      return 1;
    }
  gpprintf (0,16*10,"[Save completed]",2,-1,13);  
  return 0;
}

int LoadStateSnss(byte b)
{
  apu_t *apu;
  SNSS_FILE *f;
  char savename[255],result[256];
  byte i=0,j=255,k=255,l=255;
  int x,y,jj;
  char chaine[256];
#ifdef __asmcpu__
  uint32 nesregs[6];
#endif  

  SNSS_RETURN_CODE code;
  SNSS_BLOCK_TYPE blockType;
    
  while (CurrentROMFile[i]) 
    {
      if (CurrentROMFile[i]==92) j=i;  //92="\" 
      if (CurrentROMFile[i]==46) k=i;  //46="."
      if (CurrentROMFile[i]=='/') l=i; 
      i++;
    }
	
  CurrentROMFile[k]=0;
  if (l==255) gm_strcpy(savename,&CurrentROMFile[(byte)(j+1)]);  
  else gm_strcpy(savename,&CurrentROMFile[(byte)(l+1)]);  
  CurrentROMFile[k]=46;
  i=gm_lstrlen(savename);
  savename[i]=46;
  savename[i+1]='s';
  savename[i+2]='s';
  savename[i+3]=b%10+48;
  savename[i+4]=0;
	
  gm_sprintf(chaine,"Loading state from file : %s\n",savename);
//  DrawMessage(chaine,nflip);
  gpprintf(0,16,chaine,-1,-1,13);

  
  code = SNSS_OpenFile (&f,savename,SNSS_OPEN_READ);
  
  if (code!=SNSS_OK)
  {  
      gpprintf(0,32,"Cannot load state\n",-1,-1,13);
      return 1;
  }
          
  gm_sprintf(chaine,"Reading, got %d blocks",f->headerBlock.numberOfBlocks);
  
  gpprintf(0,32,chaine,-1,-1,13);
  /* iterate through each block in the file */
  if (f->headerBlock.numberOfBlocks>255) 
    {
      //printf("Load overflow\n"); 
      return 1;
    }
    
   // now we are too far to not reset the nes if there is a pb   
   nes_reset();
  for (j = 0; j < f->headerBlock.numberOfBlocks; j++)
    {
      if ((code = SNSS_GetNextBlockType(&blockType,f)) != SNSS_OK)
        {
	  //          printf ("Error: %s\n", getSnssErrorString (code));
          return 1;
        }
      if ((code = SNSS_ReadBlock(f,blockType)) != SNSS_OK)
        {
	  //          printf ("Error: %s\n", getSnssErrorString (code));
          return 1;
        }

      switch (blockType)
        {
        case SNSS_BASR:
          gm_sprintf(chaine,"reading base registers block\n");
	  gpprintf (0,48,chaine,-1,-1,13);
	
            
#ifndef __asmcpu__
          nes6502_getcontext(&_NEScontext);
	  _NEScontext.a_reg=f->baseBlock.regA;
	  _NEScontext.x_reg=f->baseBlock.regX;
	  _NEScontext.y_reg=f->baseBlock.regY;
	  _NEScontext.p_reg=f->baseBlock.regFlags;
	  _NEScontext.s_reg=f->baseBlock.regStack;
	  _NEScontext.pc_reg=f->baseBlock.regPc;	  
	  _NEScontext.int_pending = 0;
	  nes6502_setcontext(&_NEScontext);	  
#else	  	  	  
  	  nesregs[0]=f->baseBlock.regA;
  	  nesregs[1]=f->baseBlock.regX;
  	  nesregs[2]=f->baseBlock.regY;
  	  nesregs[3]=f->baseBlock.regFlags;  	  
  	  nesregs[4]=f->baseBlock.regPc;
  	  nesregs[5]=f->baseBlock.regStack;
  	  set_context(nesregs);  	  
#endif
	  gm_memcpy(Vnes.CPUPageIndex[0],f->baseBlock.cpuRam,0x800);	  
	  ppu_loadstate(&(f->baseBlock));
	  gm_memcpy(Vnes.PPU_nametables,f->baseBlock.ppuRam,0x1000);
	  gm_memcpy(Vnes.PPU_palette,f->baseBlock.palette,0x20);

	  for (i=0;i<0x10;i++)
	    {
	      Vnes.PPU_palette[i]&=0x3f;
	      Vnes.PPU_palette[i+0x10]&=0x3f;
	    }

	  Vnes.var.ppu_mirror0=f->baseBlock.mirrorState[0]&0x03;
	  Vnes.var.ppu_mirror1=f->baseBlock.mirrorState[1]&0x03;
	  Vnes.var.ppu_mirror2=f->baseBlock.mirrorState[2]&0x03;
	  Vnes.var.ppu_mirror3=f->baseBlock.mirrorState[3]&0x03;
	  ppu_mirror(Vnes.var.ppu_mirror0,Vnes.var.ppu_mirror1,Vnes.var.ppu_mirror2,Vnes.var.ppu_mirror3);  	  
	  
	  
	  	  	  
          break;

        case SNSS_VRAM:
          gm_sprintf(chaine,"reading VRAM block, %d\n",f->vramBlock.vramSize);
	  gpprintf (0,64,chaine,-1,-1,13);          
	  gm_memcpy(Vnes.PPU_patterntables,(f->vramBlock.vram),f->vramBlock.vramSize);	
          break;

        case SNSS_SRAM:
	  //gpprintf (0,16*5,"reading SRAM block\n",-1,-1,13);          
	  gm_sprintf(chaine,"reading SRAM block, %d\n",f->sramBlock.sramSize);
	  gpprintf (0,16*5,chaine,-1,-1,13);          
	  gm_memcpy(Vnes.NESSRAM/*Vnes.CPUMemory+0x6000*/,(f->sramBlock.sram),f->sramBlock.sramSize);
	  Vnes.var.sramEnabled=f->sramBlock.sramEnabled;
          break;

        case SNSS_MPRD:
	  gpprintf (0,16*6,"reading mapper data block\n",-1,-1,13);          	  	  
	  
          for (i = 0; i < 4; i ++)
            {
	      Vnes.CPUPageIndex[i+4]=Vnes.var.prg_beg+((uint32)f->mapperBlock.prgPages[i]<<13);
            }          
	  // set CHR pages
	  for(i = 0; i < 8; i++)
	  {
		if(f->mapperBlock.chrPages[i] & 0x8000)
		{
			// VRAM
			Vnes.PPUPageIndex[i] = Vnes.PPU_patterntables + ((uint32)(f->mapperBlock.chrPages[i] & 0x1f) << 10);
			Vnes.PPUPageIndexProtect[i]=0;
		}
		else
		{
			// VROM
			Vnes.PPUPageIndex[i] = Vnes.var.chr_beg + ((uint32)(f->mapperBlock.chrPages[i] & 0x1f) << 10);
			Vnes.PPUPageIndexProtect[i]=1;			
		}
	  }	  	  	  	  
          if (mapper->intf->set_state) mapper->intf->set_state(&(f->mapperBlock));                              
                              
          #ifdef __asmcpu__
          set_cpu_bank_full(Vnes.CPUPageIndex[4],Vnes.CPUPageIndex[5],Vnes.CPUPageIndex[6],Vnes.CPUPageIndex[7]);
          #else
          nes6502_getcontext(&_NEScontext);
          nes6502_setcontext(&_NEScontext);    
          #endif                   
                             
          break;
        case SNSS_CNTR:
	  gpprintf (0,16*7,"reading controllers block\n",-1,-1,13);          
          break;

        case SNSS_SOUN:
	  gpprintf (0,16*8,"reading sound block\n",-1,-1,13);          
            //ResetSoundState();
            apu=apu_getcontext();
          for (i = 0; i < 0x16; i ++)
            {
	      if(i == 0x14) continue;
	      Vnes.CPUMemory[0x4000+i]=f->soundBlock.soundRegisters[i];
    		// write the DMC regs directly
    		if((i >= 0x10) && (i <= 0x13))
    		{
      			apu->apus.dmc.regs[i - 0x10] = f->soundBlock.soundRegisters[i];
    		}
    		else
    		{
      			apu_write(0x4000 + i, f->soundBlock.soundRegisters[i]);
      			apu_write_cur(0x4000 + i, f->soundBlock.soundRegisters[i]);
    		}
            }                        	  	  
                        
          break;

        default:
	  //          printf ("invalid block type\n");
          break;
        }
    }

  /* close the files */
  if ((code = SNSS_CloseFile (&f)) != SNSS_OK)
    {
      //      printf ("Error: %s\n", getSnssErrorString (code));
      return 1;
    }
  gpprintf (0,16*10,"[Load completed]\n",2,-1,13);    
  
  return 0;
}

void Reset()
{
  Init_NES(CurrentROMFile);
}

void Yo_Printc(uint32 x,uint32 y,char c,uint8 *scr,byte col)
{
  int indicex,indicey;  
  uint32 i,j;
  int indice=c-32; //index in the font.h array
  

  if (indice>=96) indice=0;
  if (indice<0) indice=0;
  indicey=(indice>>4)*9;
  indicex=(indice&0xf)<<3;
  
  for (i=0;i<8;i++)
    for (j=0;j<9;j++) if (font[j+indicey][i+indicex]=='#') ((Uint8 *)(scr))[255-(y+j)+((i+x)<<8)]=col;
}


Uint8 ProcessKey()
{
  //  Uint8 *keys;
  int x,y;
  byte bouton;
  int ExKey;
  unsigned int pp,pw;
  int i;

  GpKeyGetEx(&ExKey);
  
  if (Vnes.var.padmode==2)
  {
  	//zapper
  	i=(ExKey&GPC_VK_FB?3:1);
  	if (ExKey&GPC_VK_LEFT) Vnes.var.zapperX-=i;
  	if (ExKey&GPC_VK_RIGHT) Vnes.var.zapperX+=i;
  	if (ExKey&GPC_VK_UP) Vnes.var.zapperY-=i;
  	if (ExKey&GPC_VK_DOWN) Vnes.var.zapperY+=i;
  	if (Vnes.var.zapperX<0) Vnes.var.zapperX=0;
  	if (Vnes.var.zapperY<0) Vnes.var.zapperY=0;
  	if (Vnes.var.zapperX>255) Vnes.var.zapperX=255;
  	if (Vnes.var.zapperY>239) Vnes.var.zapperY=239;
  }

  /*if ((ExKey & GPC_VK_FL) && (ExKey & GPC_VK_FR)) 
  {  	
  	return (INT_QUIT);
  }*/
  
  if (ExKey & GPC_VK_FL)
  {  	
  	fastforward=1;
  	GP_PauseAudio(1);
  }
  else 
  {
  	if (fastforward)
  	{
  	   fastforward=0;  	   
  	   GP_PauseAudio(0);
  	   /*if (Vnes.var.enablesound==2)
      	   {      	   	
	  	pp=((unsigned int)*play_pos);
	  	pw=((unsigned int)Vnes.var.sndbuffer_pos);
	  	pw=(pp-(unsigned int)Vnes.var.sndbuffer)/Vnes.var.buffsize;
	  	Vnes.var.sndbuffer_pos=(unsigned int)Vnes.var.sndbuffer+pw*Vnes.var.buffsize;
	   }*/	  	
	   framecount=0;
  	   framecounter=0;  
  	   beginTime=gp32_GpTickCountGet();
  	}
   }
  
  if ((ExKey & GPC_VK_FR) && (!key_pressed_delay) && ((ExKey & GPC_VK_FL) == 0)) /*FL is a security...*/
  {  	
  
/*    DrawMessage("pausing sound",0);
    gpwait(0);*/
  
  	GP_PauseAudio(1);

	GpGraphicModeSet(8, NULL);	
  	GpLcdSurfaceGet(&gpDraw[0], 0);   
  	GpLcdSurfaceGet(&gpDraw[1], 1);   
  	GpSurfaceSet(&gpDraw[0]);
	nflip=0;
	
/*    DrawMessage("entering gui",0);
    gpwait(0);*/
	
  	i=doGUI();  	

	nflip=0;
	if (Vnes.var.scaleMode<2) 
	{
		GpGraphicModeSet(8, NULL);
		GpLcdSurfaceGet(&gpDraw[0], 0);   
  		GpLcdSurfaceGet(&gpDraw[1], 1);   
  		GpSurfaceSet(&gpDraw[0]);
		Cls(13,0);
  		Cls(13,1);
  	}
	else 
	{	
		GpGraphicModeSet(16, NULL);			
  		GpLcdSurfaceGet(&gpDraw[0], 0);   
  		GpLcdSurfaceGet(&gpDraw[1], 1);   
  		GpSurfaceSet(&gpDraw[0]);  	
  		gm_memset((&gpDraw[0])->ptbuffer,0,320*240*2);
  		gm_memset((&gpDraw[1])->ptbuffer,0,320*240*2);
  		//Cls(0,0);
   		//Cls(0,1);   
   	}


  	beginTime=gp32_GpTickCountGet();
  	//starttime=GpTickCountGet();
  	framecount=0;
  	framecounter=0;  
  	key_pressed_delay=20;  	  	
  	if (i) return 1;
  	GP_PauseAudio(0);
  }
  
  if (key_pressed_delay>0) key_pressed_delay--;

  return 0;
}

INLINE byte FrameCompleted()
{
   int i,waitforframe;   
   int x,y;
   unsigned int pp,pw,dist;  
   
   if (ProcessKey()) 
   {    
	   return 1;   
   }
   
   if (Vnes.var.DrawCframe)
   {
	frameRendered++;
   	if (Vnes.var.padmode==2)
   	{
   		//Draw zapper 
   		int32 x,y,xx,yy;   		   		   		
   		for (x=Vnes.var.zapperX-4+32;x<(Vnes.var.zapperX-1+32);x++)
   		  if (x>=32) (&gpDraw[nflip])->ptbuffer[239-Vnes.var.zapperY+x*240]=32; //white
   		for (x=Vnes.var.zapperX+2+32;x<=(Vnes.var.zapperX+4+32);x++)
   		  if( x<256+32) (&gpDraw[nflip])->ptbuffer[239-Vnes.var.zapperY+x*240]=32; //white
   		  
   		for (y=Vnes.var.zapperY-4;y<Vnes.var.zapperY-1;y++)
   		  if (y>=0) (&gpDraw[nflip])->ptbuffer[239-y+(Vnes.var.zapperX+32)*240]=32; //white
   		for (y=Vnes.var.zapperY+2;y<=Vnes.var.zapperY+4;y++)
   		  if( y<240) (&gpDraw[nflip])->ptbuffer[239-y+(Vnes.var.zapperX+32)*240]=32; //white   		   	
   	}   	   	
   	
   	
   }      
   
   /*Draw some stats*/   
   if (frame==Vnes.var.fps) //(GpTickCountGet()-beginTime>=1000)
   {
   	if (showFPS)
   	{
        i=100*1000/(gp32_GpTickCountGet()-beginTime+1);
   		gm_sprintf(chaine,"%2d\n%3d",frameRendered,i);
   		gpprintf(0,0,chaine,-1,13,0xe0);
   		nflip^=1;
   		gpprintf(0,0,chaine,-1,13,0xe0);
   		nflip^=1;   		   		
   	}
   	beginTime=gp32_GpTickCountGet();
   	frame=0;
   	frameRendered=0;
   }
  
   waitforframe=0;
   if (Vnes.var.enablesound)
   {			  
      if (/*(Vnes.var.DrawCframe)&&*/(Vnes.var.enablesound==2))
      {
    	apu_process(Vnes.var.sndbuffer_pos,Vnes.var.buffsize);    	
    	sync_apu_register();
    	Vnes.var.sndbuffer_pos+=Vnes.var.buffsize<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo);
	   	if (Vnes.var.sndbuffer_pos>=Vnes.var.sndbuffer_end) Vnes.var.sndbuffer_pos-=(Vnes.var.buffsize*NBBUFFERS<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo));
    

      
      
		pp=((unsigned int)*play_pos);
		pw=((unsigned int)Vnes.var.sndbuffer_pos);
		if (pw>pp) dist=(pw-pp);
		else dist=((pw+sz)-pp);	  
		dist=dist>>(Vnes.var.Sound16bits+Vnes.var.SoundStereo);	    	  
/*		if ( dist<=Vnes.var.buffsize) 
		{
			Vnes.var.sndbuffer_pos+=4*Vnes.var.buffsize<<Vnes.var.Sound16bits;
   			if (Vnes.var.sndbuffer_pos>=Vnes.var.sndbuffer_end) Vnes.var.sndbuffer_pos-=(Vnes.var.buffsize*NBBUFFERS<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo));
		}
		else*/
		if ( dist<Vnes.var.buffsize*NBBUFFERS/4) 
		{		  		  
//			framecount++;
//			waitforframe=1;
			apu_process(Vnes.var.sndbuffer_pos,Vnes.var.buffsize);    		
			sync_apu_register();
    		Vnes.var.sndbuffer_pos+=Vnes.var.buffsize<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo);
	   		if (Vnes.var.sndbuffer_pos>=Vnes.var.sndbuffer_end) Vnes.var.sndbuffer_pos-=(Vnes.var.buffsize*NBBUFFERS<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo));
		}								
      }   


    	//Sound_Update(fastforward); 

    

      //FillAudio(Vnes.var.sndbuffer_pos,(Vnes.var.buffsize)<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo));
      //Vnes.var.sndbuffer_pos+=(Vnes.var.buffsize)<<(Vnes.var.Sound16bits+Vnes.var.SoundStereo);
      //if (Vnes.var.sndbuffer_pos==Vnes.var.sndbuffer_end) Vnes.var.sndbuffer_pos=Vnes.var.sndbuffer;		  
   }

   /*on passe a la frame suivante*/
   framecounter++;
      
   // max frames lost
   if (((int)(framecount-framecounter))>MAX_LOST_FRAME) framecount=framecounter+1; 
	 
   if (!Vnes.var.vsync)
   {
	#ifndef  __skipsynchro__
	   if ((!fastforward)&&(waitforframe==0))
	   {	  
    	 while (framecount<framecounter)
	     {
	       waitforframe=2;
	       __asm {nop;nop;nop;nop;}
         }
       }      
	#endif  
   }
   else
   {
    if ((!fastforward)&&(waitforframe==0))
    		if (framecount<framecounter) waitforframe=2;
   	if (Vnes.var.vsync&&Vnes.var.DrawCframe) 
    {
   		GpSurfaceFlip(&gpDraw[nflip]);   	
   		nflip^=1;
   	}
   	if ((!fastforward)&&(waitforframe==2))
	    while (framecount<framecounter)	__asm {nop;nop;nop;nop;}	          
   }

   if (fastforward) i=15;
   else i=fskiprate;
   
   if (i==AUTO_FSKIP)
   {
   	if (waitforframe==2) 
   	{   		
   		Vnes.var.DrawCframe=1;
   		frameskipped=0;   		
   	}
   	else 
   	{
   		Vnes.var.DrawCframe=0;
   		frameskipped++;
   	}   	
   }
   else
   {
   	// to avoid pb with game changing data periodically (for ex. : every 1frame/2).
   	fskiprate_adj+=fskiprate_adj_ratio;
   	if (fskiprate_adj>=100)
   	{
	   	frameskipped++;
	   	fskiprate_adj-=100;
   	}
   	if ((++frameskipped)>i) 
   	{
      		Vnes.var.DrawCframe=1;
      		frameskipped-=i+1;
   	}
   	else Vnes.var.DrawCframe=0;
   }
   frame++;
   
   return 0;
}

INLINE uint8 emulate_one_frameNTSC(void)
{      
  framecompleted=false;
  //vblank=true;  
  while (!framecompleted)
  {    
     curscanline=ppu_currentscanline(); 

     
  	
     Vnes.var.currentcpucycle+=Vnes.var.cpucycle;   
     Vnes.var.curdeccycle+=Vnes.var.deccycle;
     if (Vnes.var.curdeccycle>=65536)
     {
   	Vnes.var.currentcpucycle++;
   	Vnes.var.curdeccycle-=65536;
     }   
     if (Vnes.var.currentcpucycle>0) 
     {   	
   	synccycles=Vnes.var.currentcpucycle;
   	if (curscanline==241) synccycles+=CYCLES_BEFORE_NMI;
   	#ifndef __asmcpu__   	   	
   	Vnes.var.currentcpucycle-=nes6502_execute(Vnes.var.currentcpucycle);
   	if (Vnes.var.enablesound)
   	  if (sync_dmc_register(synccycles)) nes6502_pending_irq();
   	#else   
   	Vnes.var.currentcpucycle-=cpu_exec(Vnes.var.currentcpucycle);
   	if (Vnes.var.enablesound)
      	  if (sync_dmc_register(synccycles)) cpu_pending_irq();
      	#endif      
      	synccycles=0;
     }   
     
     if (mapper->intf->HSync) mapper->intf->HSync(curscanline);     
     if (Vnes.var.lazy_mode) ppu_lazy_scanlineNTSC();
	else
	  ppu_scanlineNTSC();
     

     switch (curscanline)
     {
	case 240:
		//framecompleted=true;
		if(!(Vnes.var.frame_irq_enabled & 0xC0))
		{
			#ifndef __asmcpu__			
			nes6502_pending_irq();
			#else		
			cpu_pending_irq();
			#endif
		}		
		break;
	case 241:
		//cpu vbl	
		if (mapper->intf->VSync) mapper->intf->VSync();
							
		#ifndef __asmcpu__
		Vnes.var.currentcpucycle-=nes6502_execute(CYCLES_BEFORE_NMI); //7);		
		if (ppu_vbl()) nes6502_nmi();		
		#else		
		Vnes.var.currentcpucycle-=cpu_exec(CYCLES_BEFORE_NMI); //7);
		if (ppu_vbl()) cpu_nmi();		
		#endif      		      		      		
		break;
	case 261:
		//vblank=false;
		framecompleted=true;
		break;

	}        
    
    
  }
  
/*  if (Vnes.var.enablesound)  
    {
    	//Sound_Update(fastforward); 
    	apu_process(Vnes.var.sndbuffer_pos,Vnes.var.buffsize);    	
    	sync_apu_register();
    	Vnes.var.sndbuffer_pos+=Vnes.var.buffsize<<Vnes.var.Sound16bits;
	   	if (Vnes.var.sndbuffer_pos>=Vnes.var.sndbuffer_end) Vnes.var.sndbuffer_pos=Vnes.var.sndbuffer;
    }*/
  
    	        	  
  return FrameCompleted();            
}


INLINE uint8 emulate_one_framePAL(void)
{      
  framecompleted=false;
  //vblank=true;  
  while (!framecompleted)
  {  
     curscanline=ppu_currentscanline(); 
  	
     Vnes.var.currentcpucycle+=Vnes.var.cpucycle;   
     Vnes.var.curdeccycle+=Vnes.var.deccycle;
     if (Vnes.var.curdeccycle>=65536)
     {
   	Vnes.var.currentcpucycle++;
   	Vnes.var.curdeccycle-=65536;
     }   
     if (Vnes.var.currentcpucycle>0) 
     {   	
   	synccycles=Vnes.var.currentcpucycle;
   	if (curscanline==241) synccycles+=CYCLES_BEFORE_NMI;
   	#ifndef __asmcpu__   	   	
   	Vnes.var.currentcpucycle-=nes6502_execute(Vnes.var.currentcpucycle);
   	if (Vnes.var.enablesound)
   	  if (sync_dmc_register(synccycles)/*&&DPCM_IRQ*/) nes6502_pending_irq();
   	#else   
   	Vnes.var.currentcpucycle-=cpu_exec(Vnes.var.currentcpucycle);
   	if (Vnes.var.enablesound)
      	  if (sync_dmc_register(synccycles)/*&&DPCM_IRQ*/) cpu_pending_irq();
      	#endif      
      	synccycles=0;
     }   
     
     if (mapper->intf->HSync) mapper->intf->HSync(curscanline);	                 
     if (Vnes.var.lazy_mode) ppu_lazy_scanlinePAL();
	 else ppu_scanlinePAL();
     switch (curscanline)
     {
	case 240:
		//framecompleted=true;
		if(!(Vnes.var.frame_irq_enabled & 0xC0))
		{
			#ifndef __asmcpu__			
			nes6502_pending_irq();
			#else		
			cpu_pending_irq();
			#endif
		}		
		break;
	case 241:
		//cpu vbl	
		if (mapper->intf->VSync) mapper->intf->VSync();
							
		#ifndef __asmcpu__
		Vnes.var.currentcpucycle-=nes6502_execute(CYCLES_BEFORE_NMI); //7);		
		if (ppu_vbl()) nes6502_nmi();		
		#else		
		Vnes.var.currentcpucycle-=cpu_exec(CYCLES_BEFORE_NMI); //7);
		if (ppu_vbl()) cpu_nmi();		
		#endif      		      		      		
		break;
	case 311:
		//vblank=false;
		framecompleted=true;
		break;
	/*default:
		break;*/
	}
        
    
    
  }
/*  if (Vnes.var.enablesound)  
    {
    	//Sound_Update(fastforward); 
    	apu_process(Vnes.var.sndbuffer_pos,Vnes.var.buffsize);    	
    	sync_apu_register();    	    	    	    	
    	Vnes.var.sndbuffer_pos+=Vnes.var.buffsize<<Vnes.var.Sound16bits;
   	if (Vnes.var.sndbuffer_pos>=Vnes.var.sndbuffer_end) Vnes.var.sndbuffer_pos=Vnes.var.sndbuffer;
    }*/
  
    	        	  
  return FrameCompleted();            
}


void ShowSplash()
{
  int i,j,fd,r,v,b,col,nflip;
  GP_HPALETTE h_logopal;
  GP_PALETTEENTRY*	tmp_entry;  
  tmp_entry = (GP_PALETTEENTRY*)gm_malloc((sizeof(GP_PALETTEENTRY) * 256));      
  
  //8bit mode setting
  GpGraphicModeSet(8, NULL);
  GpLcdSurfaceGet(&gpDraw[0], 0);   
  GpLcdSurfaceGet(&gpDraw[1], 1);   
  GpSurfaceSet(&gpDraw[0]);  
  
  gm_memset(tmp_entry,0xFFFF<<1,256*2);
  h_logopal = GpPaletteCreate(LJ_logo_palnb, tmp_entry);
  h_orgpal = GpPaletteSelect(h_logopal);
  GpPaletteRealize();
  
  gm_memcpy((&gpDraw[0])->ptbuffer,LJ_logo,320*240);
  gm_memcpy((&gpDraw[1])->ptbuffer,LJ_logo,320*240);
  nflip=0;
  
  fd=31;    
  while (fd>=0)
  { 
    for (i=0;i<256;i++)
    {
    	r=(LJ_logo_Pal[i]>>11)+fd;
    	v=((LJ_logo_Pal[i]>>6)&0x1F)+fd;
    	b=((LJ_logo_Pal[i]>>1)&0x1F)+fd;
    	if (r>0x1F) r = 0x1F;
    	if (v>0x1F) v = 0x1F;
    	if (b>0x1F) b = 0x1F;
        tmp_entry[i]=(r<<11)|(v<<6)|b<<1;
    }
    //GpPaletteDelete(h_logopal);
    h_logopal = GpPaletteCreate(LJ_logo_palnb, tmp_entry);
    GpPaletteDelete(GpPaletteSelect(h_logopal));
    //GpSurfaceFlip(&gpDraw[nflip]);
    GpPaletteRealize();            
    gpdelay(5);
    fd-=1;    
    nflip^=1;
  }  
  
  gpdelay(500);
  
  fd=0;
  while (fd<32)
  {      
    for (i=0;i<256;i++)
    {
    	r=(LJ_logo_Pal[i]>>11)+fd;
    	v=((LJ_logo_Pal[i]>>6)&0x1F)+fd;
    	b=((LJ_logo_Pal[i]>>1)&0x1F)+fd;
    	if (r>0x1F) r = 0x1F;
    	if (v>0x1F) v = 0x1F;
    	if (b>0x1F) b = 0x1F;
        tmp_entry[i]=(r<<11)|(v<<6)|b<<1;
    }
    //GpPaletteDelete(h_logopal);
    h_logopal = GpPaletteCreate(LJ_logo_palnb, tmp_entry);  
    GpPaletteDelete(GpPaletteSelect(h_logopal));
    //GpSurfaceFlip(&gpDraw[nflip]);    
    GpPaletteRealize();
    gpdelay(5);
    fd+=1;    
    nflip^=1;
  }  
  
  
  GpPaletteDelete(GpPaletteSelect(h_orgpal));
  GpPaletteRealize();
    
  gm_free(tmp_entry);
}

void InitDisplay()
{
  int i,j;  
  //8bit mode setting
  
  
  GpGraphicModeSet(8, NULL);
  i = GpLcdSurfaceGet(&gpDraw[0], 0);   
  i = GpLcdSurfaceGet(&gpDraw[1], 1);
  nflip=0;
  GpSurfaceSet(&gpDraw[nflip]);
      
  
  if ( !(GPC_LCD_ON_BIT & GpLcdStatusGet()) ) GpLcdEnable();   
  
  //makes the palette	
  for (i=0;i<64;i++)
    {
      normcolors[i].r=(palette[i].r);
      normcolors[i].g=(palette[i].g);
      normcolors[i].b=(palette[i].b);
      normcolors[i+64].r=(palette[i].r)|255;
      normcolors[i+64].g=(palette[i].g);
      normcolors[i+64].b=(palette[i].b);
      normcolors[i+128].r=(palette[i].r);
      normcolors[i+128].g=(palette[i].g)|255;
      normcolors[i+128].b=(palette[i].b);
      normcolors[i+64+128].r=(palette[i].r);
      normcolors[i+64+128].g=(palette[i].g);
      normcolors[i+64+128].b=(palette[i].b)|255;      
    }
    
  
  for (i=0;i<16;i++)
  {
  	menucolors[i].r=0;
  	menucolors[i].g=i<<4;
  	menucolors[i].b=0;  
  }
  menucolors[32].r=200;
  menucolors[32].g=200;
  menucolors[32].b=255;  
  menucolors[33].r=255;
  menucolors[33].g=255;
  menucolors[33].b=255;  
  
  menucolors[42].r=100;
  menucolors[42].g=100;
  menucolors[42].b=128;  
  menucolors[43].r=128;
  menucolors[43].g=128;
  menucolors[43].b=128;  
  menucolors[44].r=96;
  menucolors[44].g=96;
  menucolors[44].b=96;  
  
  menucolors[34].r=128;
  menucolors[34].g=100;
  menucolors[34].b=30;  
  menucolors[35].r=255;
  menucolors[35].g=210;
  menucolors[35].b=210;  
  
  menucolors[36].r=230;
  menucolors[36].g=210;
  menucolors[36].b=255;  
  
  menucolors[37].r=150;
  menucolors[37].g=100;
  menucolors[37].b=200;  
  
  menucolors[38].r=255;
  menucolors[38].g=210;
  menucolors[38].b=220;  
  
  menucolors[39].r=menucolors[38].r/2;
  menucolors[39].g=menucolors[38].g/2;
  menucolors[39].b=menucolors[38].b/2;  
  
  for (i=0;i<16;i++) 
  {
  	menucolors[64+i].r=i*16;
  	menucolors[64+i].g=i*16;
  	menucolors[64+i].b=i*16;
  	menucolors[64+i+16].r=255-i*16;
  	menucolors[64+i+16].g=255-i*16;
  	menucolors[64+i+16].b=255-i*16;
  }
  
  
  GP_SetPalette();
  GP_SetPal(2);
	
  Vnes.var.Vbuffer=(uint8 *)gm_malloc((256+16)*240);    
    
  gm_memset(Vnes.var.Vbuffer,0,256*240);
  
}


void nes_LoadIni()
{
  int i;
  int err_code;
  F_HANDLE f;
  config_data cd;  
  unsigned long filesize;
  unsigned long bread;
  //f=fopen("gp:\\gpetc\\ljgp32.ini", "rb");    
  GpFileGetSize("gp:\\gpetc\\ljgp32.ini", &filesize);  
  if (filesize==CONFIG_DATA_SIZE)
  {
    if (GpFileOpen("gp:\\gpetc\\ljgp32.ini", OPEN_R, &f)==SM_OK)
    {  	    	    	
    	if (GpFileRead(f, &cd,CONFIG_DATA_SIZE,&bread)==SM_OK)
    	{
    
	    	Vnes.var.enablesound=(cd.enablesound?2:0);
	    	fskiprate=cd.fskiprate;
	    	if (fskiprate<0) fskiprate=0;
	    	if ((fskiprate!=AUTO_FSKIP)&&(fskiprate>9)) fskiprate=0;
	    	fskiprate_adj_ratio=cd.fskiprate_adj_ratio;
	    	fskiprate_adj_ratio/=10;
	    	if (fskiprate_adj_ratio>5) fskiprate_adj_ratio=0;
	    	if (fskiprate_adj_ratio<0) fskiprate_adj_ratio=0;
	    	fskiprate_adj_ratio*=10;
	    	Vnes.var.morethan8spr=(cd.morethan8spr?1:0);
	    	gamma=cd.gamma;    	
	    	if ((uint32)gamma>9) gamma=0;
	    	showFPS=(cd.showFPS?1:0);
	    	Vnes.var.DrawAllLines=(cd.DrawAllLines?1:0);
	    	
	    	Vnes.var.sndfreq = cd.soundQuality;
	    	Vnes.var.sndfilter = cd.soundFilter;
	    	
	    	Vnes.var.autofireA = cd.autoFireA;
	    	Vnes.var.autofireB = cd.autoFireB;
	    	
	    	if (gm_lstrlen(cd.lastRom)) gm_strcpy(CurrentROMFile,cd.lastRom);	    	
    	}
    	GpFileClose(f);
    }    
  }
  else
  {    	
  	GpFileRemove("gp:\\gpetc\\ljgp32.ini");
  }        	
}

void nes_SaveIni()
{
  int i;
  int err_code;
  F_HANDLE f;
  config_data cd;  
   
  //f=fopen("gp:\\gpetc\\ljgp32.ini", "w+b");
  if (GpFileCreate("gp:\\gpetc\\ljgp32.ini", ALWAYS_CREATE, &f)!=SM_OK)
  {
     //gpprintf(0,48,"No sram data",5);
  }
  else
  {
    cd.version_major='0';
    cd.version_minor='3';
    cd.enablesound=Vnes.var.enablesound;
    cd.fskiprate=fskiprate;
    cd.morethan8spr=Vnes.var.morethan8spr;
    cd.showFPS=showFPS;
    cd.gamma=gamma;
    cd.DrawAllLines=Vnes.var.DrawAllLines;
    cd.fskiprate_adj_ratio=fskiprate_adj_ratio;
    
    cd.soundQuality = Vnes.var.sndfreq;
   	cd.soundFilter = Vnes.var.sndfilter;
	    	
   	cd.autoFireA = Vnes.var.autofireA;
   	cd.autoFireB = Vnes.var.autofireB;
    
    gm_strcpy(cd.lastRom,CurrentROMFile);
    GpFileWrite(f, &cd,CONFIG_DATA_SIZE);
    GpFileClose(f);
  }	
}


void nes_init()
{  
  Vnes.var.cpucycle=113;
  Vnes.var.deccycle=65536*200/3;
  CYCLES_BEFORE_NMI = 1;
  Vnes.var.curdeccycle=0;
  Vnes.var.currentcpucycle=0;
  Vnes.var.debugmode=0; 
  
  Vnes.var.morethan8spr=0;
  Vnes.var.DrawAllLines=0;
  Vnes.var.scaleMode=0;
  ppu_setscaleMode(Vnes.var.scaleMode);
  Vnes.var.autofireA=0;
  Vnes.var.autofireB=0;
  Vnes.var.padmode=0;
  Vnes.var.zapperX=128;
  Vnes.var.zapperY=120;
  
  fskiprate=AUTO_FSKIP;  //1 : on saute 1image/2
  fskiprate_adj_ratio=0;
  fskiprate_adj=0;
  Vnes.var.InterPolateSnd=0;
  Vnes.var.emupause=0;  
  Vnes.var.fps=60; //ntsc
  Vnes.var.enablesound=2;
  Vnes.var.MasterVolume=127;
  
  Vnes.var.sndfreq=1;
  Vnes.var.sndfilter=2;

  
  
  showFPS=0;
  gamma=0;
}

void CloseAll()
{      
      /*restore clock*/

      GpClockSpeedChange(67800000, 0x69032, 3); 

      /*restore sys pal*/
      GP_SetPal(2);
      /*clear screen*/
      nflip=0;
      Cls(0,0);
      Cls(0,1);  		
      GpSurfaceSet(&gpDraw[nflip]);      
      /*save settings*/      
      nes_SaveIni();
      /*stop timer*/
      GpTimerKill(0);            
      /*close soundsys*/
      Close_SoundSys();
      /*close rom*/
      Close_ROM(1);      
}

Uint8 nes_main (int runcounter) //(int argc,char *argv[])
{
  char i;
  int j,k;
  char chaine[256*2];
  FSEL filesel;  
     
  
  CurrentROMFile[0]=0;  
      
  
  stepbystep=0;
  debugison=0;
  
  nes_init();
  
  
//initializes file system
  GpFatInit();

  nes_LoadIni();
  
#ifndef __testing__
  if (!runcounter) ShowSplash();
#endif
  
  InitDisplay();
  
  

    
#ifndef __testing__  
  /*if (!runcounter)
  {  
	DrawMessage("[Little John GP32] v0.3 by Yoyo\n-------------\nLogo by Franck\n\nBased on Shatbox, Nester, LJNG, FCA\n\nPlease have a look at the\nReadme.txt\nFor more informations\n\nPlease press a button",nflip);
	gpwait(20);
  }*/
#endif  

  
  filesel.bgcolor=255;
  filesel.bgfilecolor=255;
  filesel.bgdircolor=215;
  filesel.curdircolor=4;
  filesel.dircolor=1;
  filesel.dirselcolor=0;
  filesel.filecolor=66;
  filesel.fileselcolor=64;
  filesel.fname=CurrentROMFile;
  filesel.gpDraw=gpDraw;    
  nflip=0;

  
#ifndef __testing__

  //GpClockSpeedChange(67800000, 0x69032, 3); 
  GpClockSpeedChange(132000000, 0x3a011, 3);    

  chaine[0]=0;
  if (gm_lstrlen(filesel.fname)>0)
  {
	k=0;
	j=0;
	gm_strcpy(chaine,filesel.fname);
	while (chaine[k])
	{
		if (chaine[k]=='\\') j=k;
		k++;
	}    
    chaine[j]=0;
    gm_strcpy(filesel.fname,(char*)(filesel.fname+j+1));
  } 
  
  if (chaine[0]==0) gm_sprintf(chaine,"gp:\\gpmm\\nes");
  FileRequest ( &filesel, chaine, 0);
  //gm_sprintf(CurrentROMFile,"gp:\\gpmm\\nes\\uwaters.zip");

  GpClockSpeedChange(132000000, 0x3a011, 3);    

#else
 GpRelativePathSet ( "gp:\\gpmm\\nes" );
 gm_sprintf(CurrentROMFile,"gp:\\gpmm\\nes\\dragon~4.zip");
#endif

  
  
  nflip=0;
  Cls(0,0);
  Cls(0,1);
  
  //GpSurfaceFlip(&gpDraw[nflip]);
  GpSurfaceSet(&gpDraw[nflip]);
  //	GpSurfaceFlip(&gpDraw[nflip++]);
  //	nflip &= 0x01;
  
  
  if ((!Load_ROM(CurrentROMFile))&&(!Open_ROM())&&(!Init_NES(CurrentROMFile)))
    {     
              	
      Open_SoundSys(Vnes.var.sndfreq,Vnes.var.sndfilter);
      gp32AudioPaused=1;
      /*timer*/
      GpTimerOptSet(0, (Vnes.var.fps)/__timerdiv__, 0, timertps);  // FIX: doesn't work ok...
      GpTimerSet(0);
      
      LaunchEmu();
            
      CloseAll();
      
      
      /*exit*/
      return(0);
    }
  else 
    {
      Close_ROM(0);
      return(1);
    }
}


void asm_panic(uint32 opcode,uint32 PC,uint8 *nesstack,uint8 *pcmem)
{
	int i;
	gm_sprintf(chaine,"PANIC : op %02X PC=%02X\nStack mem : %02X %02X %02X %02X\nPC mem : %02X %02X %02X %02X %02X %02X\n%02X %02X %02X %02X %02X %02X",
	opcode,PC,nesstack[0],nesstack[1],nesstack[2],nesstack[3],
	pcmem[-6],pcmem[-5],pcmem[-4],pcmem[-3],pcmem[-2],pcmem[-1],
	pcmem[0],pcmem[1],pcmem[2],pcmem[3],pcmem[4],pcmem[5]);
	gpprintf(0,0,chaine,0,0,0xe0);
			
	GpAppExit();
}


void reset_asmcpu(void)
{			
#ifdef __asmcpu__
	Vnes.NESRAM=NULL;
	init_cpu((uint8*)(&Vnes.NESRAM));	
	
	Vnes.CPUPageIndex[0]=Vnes.NESRAM;
		
	init_sram(Vnes.CPUPageIndex[3]);	
	//gm_memset(Vnes.NESSRAM,0,0x10000);
	gm_memset(Vnes.NESRAM,0,0x7ff);	
	reset_cpu();			
#endif
}

void Cdebug_opcode(uint32 *nesreg)
{
	gm_sprintf(chaineDEB,"A=%02X X=%02X Y=%02X S=%02X P=%02X\nePC=%08X PC=%04X CYCLE=%02X\nopcode=%02X",
	nesreg[0]>>24,nesreg[1],nesreg[2],nesreg[4]>>24,nesreg[5],
	nesreg[3],nesreg[10],nesreg[9],
	nesreg[11]);
	gpprintf(16,16,chaineDEB,0,0,0xe0);
}

void debug_info()
{  
  gpprintf(16,16,chaineDEB,0,0,0xe0);  
}